//******************************************************************************
//
// Copyright (c) 2015 Microsoft Corporation. All rights reserved.
//
// This code is licensed under the MIT License (MIT).
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//******************************************************************************

#import "Foundation/Foundation.h"
#import "Starboard.h"
#import "TestFramework.h"

#import <objc/objc-arc.h>

// Helper for comparing NSMapTables
void assertMapTablesEqual(NSMapTable* table1, NSMapTable* table2) {
    ASSERT_EQ([table1 count], [table2 count]);
    for (id key in table1) {
        ASSERT_OBJCEQ([table1 objectForKey:key], [table2 objectForKey:key]);
    }
}

TEST(NSMapTable, StrongStrong) {
    StrongId<NSMapTable> mapTable = [NSMapTable strongToStrongObjectsMapTable];
    NSObject* key1 = [NSObject new];
    NSObject* key2 = [NSObject new];
    NSObject* key3 = [NSObject new];
    NSObject* value1 = [NSObject new];
    NSObject* value2 = [NSObject new];
    NSObject* value3 = [NSObject new];

    [mapTable setObject:value1 forKey:key1];
    [mapTable setObject:value2 forKey:key2];
    [mapTable setObject:value3 forKey:key3];

    // Check current state of the table
    ASSERT_EQ(3, [mapTable count]);
    ASSERT_OBJCEQ(value1, [mapTable objectForKey:key1]);
    ASSERT_OBJCEQ(value2, [mapTable objectForKey:key2]);
    ASSERT_OBJCEQ(value3, [mapTable objectForKey:key3]);

    [key1 release]; // release just key for pair 1
    [value2 release]; // release just value for pair 2
    [key3 release]; // release both key and value for pair 3
    [value3 release];

    // State should remain the same due to table's retains
    ASSERT_EQ(3, [mapTable count]);
    ASSERT_OBJCEQ(value1, [mapTable objectForKey:key1]);
    ASSERT_OBJCEQ(value2, [mapTable objectForKey:key2]);
    ASSERT_OBJCEQ(value3, [mapTable objectForKey:key3]);
    ASSERT_TRUE([value1 retainCount] > 0);
    ASSERT_TRUE([value2 retainCount] > 0);
    ASSERT_TRUE([value3 retainCount] > 0);

    // Release the rest as cleanup;
    [value1 release];
    [key2 release];
}

TEST(NSMapTable, WeakValue) {
    StrongId<NSMapTable> mapTable = [NSMapTable strongToWeakObjectsMapTable];
    NSObject* objKey = [NSObject new];
    NSObject* objValue = [NSObject new];

    [mapTable setObject:objValue forKey:objKey];
    ASSERT_NE(nil, [mapTable objectForKey:objKey]);
    ASSERT_EQ(objValue, [mapTable objectForKey:objKey]);

    // Verify that the stored value is autonilled on last release
    [objValue release];
    ASSERT_EQ(nil, [mapTable objectForKey:objKey]);

    [objKey release];
}

TEST(NSMapTable, WeakKey) {
    StrongId<NSMapTable> mapTable = [NSMapTable weakToStrongObjectsMapTable];
    NSObject* objKey = [NSObject new];
    NSObject* objValue = [NSObject new];

    // check that count is accurate
    [mapTable setObject:objValue forKey:objKey];
    ASSERT_EQ(objValue, [mapTable objectForKey:objKey]);
    ASSERT_EQ(1, [mapTable count]);

    // Verify that the stored entry is automatically removed when the last ref to the key is released
    [objKey release];
    ASSERT_EQ(0, [mapTable count]);

    [objValue release];
}

TEST(NSMapTable, MultipleWeakKey) {
    StrongId<NSMapTable> mapTable = [NSMapTable weakToStrongObjectsMapTable];
    NSObject* objKey1 = [NSObject new];
    NSObject* objKey2 = [NSObject new];
    NSObject* objKey3 = [NSObject new];
    NSObject* objValue = [NSObject new];

    [mapTable setObject:objValue forKey:objKey1];
    [mapTable setObject:objValue forKey:objKey2];
    [mapTable setObject:objValue forKey:objKey3];
    ASSERT_EQ(objValue, [mapTable objectForKey:objKey1]);
    ASSERT_EQ(objValue, [mapTable objectForKey:objKey2]);
    ASSERT_EQ(objValue, [mapTable objectForKey:objKey3]);
    ASSERT_EQ(3, [mapTable count]);

    [objKey1 release];
    [objKey2 release];
    [objKey3 release];
    ASSERT_EQ(0, [mapTable count]);

    [objValue release];
}

TEST(NSMapTable, WeakChain) {
    StrongId<NSMapTable> mapTable = [NSMapTable weakToStrongObjectsMapTable];
    NSObject* obj1 = [NSObject new];
    NSObject* obj2 = [NSObject new];
    NSObject* obj3 = [NSObject new];
    NSObject* obj4 = [NSObject new];
    NSObject* obj5 = [NSObject new];

    [mapTable setObject:obj2 forKey:obj1];
    [mapTable setObject:obj3 forKey:obj2];
    [mapTable setObject:obj4 forKey:obj3];
    [mapTable setObject:obj5 forKey:obj4];

    [obj2 release];
    [obj3 release];
    [obj4 release];
    [obj5 release];
    ASSERT_EQ(4, [mapTable count]);

    [obj1 release];
    ASSERT_EQ(0, [mapTable count]);
}

TEST(NSMapTable, WeakWeak) {
    StrongId<NSMapTable> mapTable = [NSMapTable weakToWeakObjectsMapTable];
    NSObject* key1 = [NSObject new];
    NSObject* key2 = [NSObject new];
    NSObject* key3 = [NSObject new];
    NSObject* value1 = [NSObject new];
    NSObject* value2 = [NSObject new];
    NSObject* value3 = [NSObject new];

    [mapTable setObject:value1 forKey:key1];
    [mapTable setObject:value2 forKey:key2];
    [mapTable setObject:value3 forKey:key3];
    ASSERT_EQ(3, [mapTable count]);

    [value1 release];
    [value2 release];
    ASSERT_EQ(3, [mapTable count]);
    ASSERT_OBJCEQ(nil, [mapTable objectForKey:key1]);
    ASSERT_OBJCEQ(nil, [mapTable objectForKey:key2]);
    ASSERT_OBJCEQ(value3, [mapTable objectForKey:key3]);

    [key1 release];
    [key2 release];
    ASSERT_EQ(1, [mapTable count]);
    ASSERT_OBJCEQ(value3, [mapTable objectForKey:key3]);

    [key3 release];
    ASSERT_EQ(0, [mapTable count]);

    [value3 release];
}

TEST(NSMapTable, RemoveObject) {
    StrongId<NSMapTable> mapTable = [NSMapTable strongToStrongObjectsMapTable];
    NSObject* key = [NSObject new];
    NSObject* value = [NSObject new];

    // use objc_storeWeak to monitor whether the pointers have been released
    NSObject* keyStore;
    NSObject* valueStore;
    objc_storeWeak(&keyStore, key);
    objc_storeWeak(&valueStore, value);

    [mapTable setObject:value forKey:key];
    ASSERT_EQ(1, [mapTable count]);

    [key release];
    [value release];
    ASSERT_EQ(1, [mapTable count]);
    ASSERT_OBJCNE(nil, keyStore);
    ASSERT_OBJCNE(nil, valueStore);

    [mapTable removeObjectForKey:key];
    ASSERT_EQ(0, [mapTable count]);
    ASSERT_OBJCEQ(nil, keyStore);
    ASSERT_OBJCEQ(nil, valueStore);
}

TEST(NSMapTable, RemoveAllObjects) {
    StrongId<NSMapTable> mapTable = [NSMapTable strongToStrongObjectsMapTable];
    NSObject* key1 = [NSObject new];
    NSObject* key2 = [NSObject new];
    NSObject* key3 = [NSObject new];
    NSObject* value1 = [NSObject new];
    NSObject* value2 = [NSObject new];
    NSObject* value3 = [NSObject new];

    // use objc_storeWeak to monitor whether the pointers have been released
    NSObject* keyStore;
    NSObject* valueStore;
    objc_storeWeak(&keyStore, key1);
    objc_storeWeak(&valueStore, value1);

    [mapTable setObject:value1 forKey:key1];
    [mapTable setObject:value2 forKey:key2];
    [mapTable setObject:value3 forKey:key3];
    ASSERT_EQ(3, [mapTable count]);

    [key1 release];
    [value1 release];
    ASSERT_EQ(3, [mapTable count]);
    ASSERT_OBJCNE(nil, keyStore);
    ASSERT_OBJCNE(nil, valueStore);

    [mapTable removeAllObjects];
    ASSERT_EQ(0, [mapTable count]);
    ASSERT_OBJCEQ(nil, keyStore);
    ASSERT_OBJCEQ(nil, valueStore);

    // clean up
    [key2 release];
    [key3 release];
    [value2 release];
    [value3 release];
}

TEST(NSMapTable, SameKey) {
    StrongId<NSMapTable> mapTable = [NSMapTable weakToStrongObjectsMapTable];
    NSObject* key1 = [NSObject new];
    NSObject* value1 = [NSObject new];
    NSObject* value2 = [NSObject new];
    NSObject* value3 = [NSObject new];

    // use objc_storeWeak to monitor whether the pointers have been released
    NSObject* valueStore1;
    NSObject* valueStore2;
    NSObject* valueStore3;
    objc_storeWeak(&valueStore1, value1);
    objc_storeWeak(&valueStore2, value2);
    objc_storeWeak(&valueStore3, value3);

    [mapTable setObject:value1 forKey:key1];
    [value1 release];
    ASSERT_OBJCNE(nil, valueStore1);
    ASSERT_OBJCNE(nil, valueStore2);
    ASSERT_OBJCNE(nil, valueStore3);
    ASSERT_OBJCEQ(valueStore1, [mapTable objectForKey:key1]);
    ASSERT_EQ(1, [mapTable count]);

    // Should release the last ref to value1
    [mapTable setObject:value2 forKey:key1];
    [value2 release];
    ASSERT_OBJCEQ(nil, valueStore1);
    ASSERT_OBJCNE(nil, valueStore2);
    ASSERT_OBJCNE(nil, valueStore3);
    ASSERT_OBJCEQ(valueStore2, [mapTable objectForKey:key1]);
    ASSERT_EQ(1, [mapTable count]);

    // Should release the last ref to value2
    [mapTable setObject:value3 forKey:key1];
    [value3 release];
    ASSERT_OBJCEQ(nil, valueStore1);
    ASSERT_OBJCEQ(nil, valueStore2);
    ASSERT_OBJCNE(nil, valueStore3);
    ASSERT_OBJCEQ(valueStore3, [mapTable objectForKey:key1]);
    ASSERT_EQ(1, [mapTable count]);

    // Should remove the last entry and release the last ref to value3
    [key1 release];
    ASSERT_EQ(0, [mapTable count]);
    ASSERT_OBJCEQ(nil, valueStore3);
}

TEST(NSMapTable, Enumerators) {
    StrongId<NSMapTable> mapTable = [NSMapTable strongToStrongObjectsMapTable];
    StrongId<NSArray> keySet = @[ @-1, @-2, @-3, @-4, @-5 ];
    StrongId<NSArray> valueSet = @[ @1, @2, @3, @4, @5 ];

    for (size_t i = 0; i < [keySet count]; i++) {
        [mapTable setObject:[valueSet objectAtIndex:i] forKey:[keySet objectAtIndex:i]];
    }

    // keys, allObjects
    {
        StrongId<NSEnumerator> enumerator = [mapTable keyEnumerator];
        ASSERT_OBJCEQ(keySet, [enumerator allObjects]);
        ASSERT_OBJCEQ(nil, [enumerator nextObject]);
    }

    // keys, nextObject
    {
        StrongId<NSEnumerator> enumerator = [mapTable keyEnumerator];
        for (size_t i = 0; i < [keySet count]; i++) {
            ASSERT_OBJCEQ([keySet objectAtIndex:i], [enumerator nextObject]);
        }
        ASSERT_OBJCEQ(nil, [enumerator nextObject]);
    }

    // values, allObjects
    {
        StrongId<NSEnumerator> enumerator = [mapTable objectEnumerator];
        ASSERT_OBJCEQ(valueSet, [enumerator allObjects]);
        ASSERT_OBJCEQ(nil, [enumerator nextObject]);
    }

    // values, nextObject
    {
        StrongId<NSEnumerator> enumerator = [mapTable objectEnumerator];
        for (size_t i = 0; i < [valueSet count]; i++) {
            ASSERT_OBJCEQ([valueSet objectAtIndex:i], [enumerator nextObject]);
        }
        ASSERT_OBJCEQ(nil, [enumerator nextObject]);
    }
}

TEST(NSMapTable, FastEnumeration) {
    StrongId<NSMapTable> mapTable = [NSMapTable strongToStrongObjectsMapTable];
    StrongId<NSArray> keySet =
        @[ @"1", @"2", @"3", @"4", @"5", @"6", @"7", @"8", @"9", @"10", @"11", @"12", @"13", @"14", @"15", @"16", @"17" ];
    StrongId<NSArray> valueSet =
        @[ @"I", @"II", @"III", @"IV", @"V", @"VI", @"VII", @"VIII", @"IX", @"X", @"XI", @"XII", @"XIII", @"XIV", @"XV", @"XVI", @"XVII" ];

    for (size_t i = 0; i < [keySet count]; i++) {
        [mapTable setObject:[valueSet objectAtIndex:i] forKey:[keySet objectAtIndex:i]];
    }

    StrongId<NSMutableSet> keysNotIterated = [NSMutableSet setWithArray:keySet];
    ASSERT_EQ([keySet count], [keysNotIterated count]);
    NSMapTable* pMapTable = &*mapTable;
    for (id obj in pMapTable) { // compiles to countByEnumeratingWithState
        [keysNotIterated removeObject:obj];
    }
    ASSERT_EQ(0, [keysNotIterated count]);
}

TEST(NSMapTable, DictionaryRepresentation) {
    StrongId<NSMapTable> mapTable = [NSMapTable weakToWeakObjectsMapTable];
    StrongId<NSArray> keySet = @[ @":(", @":[", @".-.", @":<" ];
    StrongId<NSArray> valueSet = @[ @":)", @":]", @"^-^", @":>" ];

    for (size_t i = 0; i < [keySet count]; i++) {
        [mapTable setObject:[valueSet objectAtIndex:i] forKey:[keySet objectAtIndex:i]];
    }

    StrongId<NSDictionary> dict = [mapTable dictionaryRepresentation];
    ASSERT_EQ([mapTable count], [dict count]);
    for (size_t i = 0; i < [keySet count]; i++) {
        id key = [keySet objectAtIndex:i];
        ASSERT_OBJCEQ([mapTable objectForKey:key], [dict objectForKey:key]);
    }
}

TEST(NSMapTable, Copy) {
    StrongId<NSMapTable> mapTable = [NSMapTable weakToWeakObjectsMapTable];
    NSObject* key1 = [NSObject new];
    NSObject* key2 = [NSObject new];
    NSObject* key3 = [NSObject new];
    NSObject* value1 = [NSObject new];
    NSObject* value2 = [NSObject new];
    NSObject* value3 = [NSObject new];

    [mapTable setObject:value1 forKey:key1];
    [mapTable setObject:value2 forKey:key2];
    [mapTable setObject:value3 forKey:key3];

    StrongId<NSMapTable> mapTableCopy = [[mapTable copy] autorelease];
    assertMapTablesEqual(mapTable, mapTableCopy);

    [value1 release];
    [value2 release];
    assertMapTablesEqual(mapTable, mapTableCopy);
    ASSERT_OBJCEQ(nil, [mapTableCopy objectForKey:key1]);
    ASSERT_OBJCEQ(nil, [mapTableCopy objectForKey:key2]);

    [key1 release];
    [key2 release];
    assertMapTablesEqual(mapTable, mapTableCopy);
    ASSERT_EQ(1, [mapTableCopy count]);

    [key3 release];
    assertMapTablesEqual(mapTable, mapTableCopy);
    ASSERT_EQ(0, [mapTableCopy count]);

    [value3 release];
}

TEST(NSMapTable, EncodeDecode) {
    StrongId<NSMapTable> mapTable = [NSMapTable weakToWeakObjectsMapTable];
    [mapTable setObject:@"value2" forKey:@"key1"];
    [mapTable setObject:@"value2" forKey:@"key2"];

    StrongId<NSMapTable> mapTableDecoded =
        [NSKeyedUnarchiver unarchiveObjectWithData:[NSKeyedArchiver archivedDataWithRootObject:mapTable]];
    assertMapTablesEqual(mapTable, mapTableDecoded);
}

NSUInteger _AlwaysSameHashFunction(const void* item, NSUInteger (*size)(const void* item)) {
    return 234324; // arbitrary
}

BOOL _AlwaysSameEqualFunction(const void* item1, const void* item2, NSUInteger (*size)(const void* item)) {
    return YES;
}

TEST(NSMapTable, CustomPointerFunctions) {
    // Verify that custom hash/isEqual functions get used by choosing very silly pointer functions
    StrongId<NSPointerFunctions> pointerFunctions = [[[NSPointerFunctions alloc] initWithOptions:0] autorelease];
    [pointerFunctions setHashFunction:&_AlwaysSameHashFunction];
    [pointerFunctions setIsEqualFunction:&_AlwaysSameEqualFunction];

    StrongId<NSMapTable> mapTable =
        [[[NSMapTable alloc] initWithKeyPointerFunctions:pointerFunctions valuePointerFunctions:pointerFunctions capacity:0] autorelease];
    NSObject* key1 = [NSObject new];
    NSObject* key2 = [NSObject new];
    NSObject* value1 = [NSObject new];
    NSObject* value2 = [NSObject new];

    [mapTable setObject:value1 forKey:key1];
    ASSERT_EQ(1, [mapTable count]);
    ASSERT_OBJCEQ(value1, [mapTable objectForKey:key1]);
    // Since key1 == key2 from the point of view of the above functions...
    ASSERT_OBJCEQ(value1, [mapTable objectForKey:key2]);

    // If the above functions got used, a hash collision should be triggered, and since the key is equal, the entry will be replaced
    [mapTable setObject:value2 forKey:key2];

    ASSERT_EQ(1, [mapTable count]);
    ASSERT_OBJCEQ(value2, [mapTable objectForKey:key2]);
    ASSERT_OBJCEQ(value2, [mapTable objectForKey:key1]);

    [key1 release];
    [key2 release];
    [value1 release];
    [value2 release];
}

TEST(NSMapTable, CopyIn) {
    StrongId<NSMapTable> mapTable =
        [[[NSMapTable alloc] initWithKeyOptions:NSMapTableStrongMemory valueOptions:(NSMapTableStrongMemory | NSMapTableCopyIn) capacity:1]
            autorelease];

    StrongId<NSObject> key = [[NSObject new] autorelease];
    StrongId<NSCalendar> value = [NSCalendar currentCalendar];

    [mapTable setObject:value forKey:key];

    ASSERT_NE(&*value, [mapTable objectForKey:key]);
}

TEST(NSMapTable, ObjectPointerPersonality) {
    StrongId<NSMapTable> mapTable =
        [NSMapTable mapTableWithKeyOptions:NSMapTableObjectPointerPersonality valueOptions:NSMapTableObjectPointerPersonality];
    // Add/remove and make sure values are as expected
    StrongId<NSArray> keySet =
        @[ [[NSObject new] autorelease], [[NSObject new] autorelease], [[NSObject new] autorelease], [[NSObject new] autorelease] ];
    StrongId<NSArray> valueSet =
        @[ [[NSObject new] autorelease], [[NSObject new] autorelease], [[NSObject new] autorelease], [[NSObject new] autorelease] ];

    for (size_t i = 0; i < [keySet count]; i++) {
        [mapTable setObject:[valueSet objectAtIndex:i] forKey:[keySet objectAtIndex:i]];
    }

    ASSERT_EQ([keySet count], [mapTable count]);

    for (size_t i = 0; i < [keySet count]; i++) {
        ASSERT_OBJCEQ([valueSet objectAtIndex:i], [mapTable objectForKey:[keySet objectAtIndex:i]]);
    }

    [mapTable removeAllObjects];
    ASSERT_EQ(0, [mapTable count]);

    // Test key equality
    [mapTable setObject:[valueSet objectAtIndex:0] forKey:[keySet objectAtIndex:0]];
    [mapTable setObject:[valueSet objectAtIndex:1] forKey:[keySet objectAtIndex:0]];
    ASSERT_EQ(1, [mapTable count]);
}